/*         ______   ___    ___ 
 *        /\  _  \ /\_ \  /\_ \ 
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      Low-level file hooking mechanism.
 *
 *      By Ben Davis.
 *
 *      See readme.txt for copyright information.
 */


#include <stdio.h>

#include "allegro.h"
#include "allegro/internal/aintern.h"

#ifndef ALLEGRO_MPW
   #include <sys/stat.h>
#endif



/* permissions to use when opening files */
#ifndef ALLEGRO_MPW
#ifdef S_IRGRP
#define OPEN_PERMS   (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)
#else
#define OPEN_PERMS   (S_IRUSR | S_IWUSR)
#endif
#endif



/* al_open():
 *  Wrapper around open(). In addition to accepting a Unicode filename which
 *  will be converted to ASCII, this function converts the path into a native
 *  format when used on the Mac.
 *
 *  WARNING: The file descriptor returned is not compatible with the
 *  al_filehook_*() functions. See al_virtualise_fd().
 */
int al_open(AL_CONST char *filename, int mode)
{
   char tmp[1024];

#ifndef ALLEGRO_MPW
   return open(uconvert_toascii(filename, tmp), mode, OPEN_PERMS);
#else
   return _al_open(uconvert_toascii(filename, tmp), mode);
#endif
}



static int default_close(void *data)
{
   int rv = close(*(int *)data);
   free(data);
   *allegro_errno = errno;
   return rv;
}

static long default_lseek(void *data, long offset, int whence)
{
   int rv = lseek(*(int *)data, offset, whence);
   *allegro_errno = errno;
   return rv;
}

static long default_read(void *data, void *buffer, long length)
{
   int rv = read(*(int *)data, buffer, length);
   *allegro_errno = errno;
   return rv;
}

static int default_write(void *data, AL_CONST void *buffer, long count)
{
   int rv = write(*(int *)data, buffer, count);
   *allegro_errno = errno;
   return rv;
}



static AL_FILEHOOKS default_filehooks = {
   &default_close, &default_lseek, &default_read, &default_write
};



#define MAX_OPEN_HOOKS 32

static int n_open_hooks = 0;
static AL_OPEN_HOOK open_hook[MAX_OPEN_HOOKS];



/* al_add_open_hook:
 *  Adds an open hook function to the list. Whenever a physical file is
 *  opened (by calling al_filehook_open()), these will be called in reverse
 *  order, and each function will have the opportunity to hook the file. If
 *  no open hook hooks the file, the default hooks will be used; these simply
 *  chain to the libc functions.
 *
 *  When hooking a file, you may use malloc() to allocate memory for any data
 *  you need on this file. Store the pointer using the 'data' parameter:
 *
 *     MYSTRUCT *mydata = (MYSTRUCT *) (*data = malloc(sizeof(MYSTRUCT)));
 *     mydata->fd = ...
 *
 *  Then return a pointer to an AL_FILEHOOKS struct, which must be global or
 *  static (it must continue to exist after your function returns). This
 *  struct contains pointers to your functions, and the pointer you stored in
 *  *data will be passed to each of these. In particular, your close function
 *  should free up the data.
 *
 *  You do not have to use the data pointer at all if you don't want to; just
 *  leave it uninitialised.
 *
 *  If you do not wish to hook the file, return NULL. If you have called any
 *  functions that may have set errno (note that even successful function
 *  calls will set errno sometimes), then you should set errno to 0 before
 *  returning.
 *
 *  If you did wish to hook this file but an error occurred, set errno to an
 *  appropriate value and return NULL. The file open request will fail.
 *
 *  al_add_open_hook() can store up to 32 function pointers. It will return
 *  nonzero if you try to add too many. Trying to add a function twice will
 *  have no effect and return zero. If there is any risk of your function
 *  being called recursively, you must set up your own semaphore system in
 *  order to avoid hooking the file on the second recursive call. This could
 *  get complicated if thread safety is important, which is why Allegro
 *  doesn't provide a semaphore itself.
 */
int al_add_open_hook(AL_OPEN_HOOK hook)
{
   int i;

   ASSERT(hook);

   for (i = 0; i < n_open_hooks; i++)
      if (open_hook[i] == hook)
	 return 0;

   if (n_open_hooks >= MAX_OPEN_HOOKS)
      return 1;

   open_hook[n_open_hooks++] = hook;
   return 0;
}



/* al_remove_open_hook:
 *  Removes an open hook from the list. Pass NULL to remove all hooks.
 */
void al_remove_open_hook(AL_OPEN_HOOK hook)
{
   if (hook) {
      int i;
      for (i = 0; i < n_open_hooks; i++) {
	 if (open_hook[i] == hook) {
	    n_open_hooks--;
	    while (i < n_open_hooks) {
	       open_hook[i] = open_hook[i+1];
	       i++;
	    }
	    return;
	 }
      }
   } else
      n_open_hooks = 0;
}



#define MAX_FILES 256

static int n_files = 0;
static AL_FILEHOOKS *filehooks[MAX_FILES];
static void *filedata[MAX_FILES];

/* When a file handle is dup'd, the descriptors are linked in a double
 * circular list. filedupnext[] and filedupprev[] index the next and previous
 * file descriptors. When a file descriptor is closed, the file itself is
 * only closed if filedupnext[] points to the same file descriptor.
 */
static unsigned char filedupnext[MAX_FILES];
static unsigned char filedupprev[MAX_FILES];



/* al_virtualise_fd:
 *  Takes a real file descriptor, perhaps returned by al_open(), and converts
 *  it into a virtual one suitable for use with the al_filehook_*()
 *  functions. If it succeeds, you will no longer need to worry about the
 *  real file descriptor; it will get closed for you when you close the
 *  virtual file descriptor (use dup() to get around this). On the other
 *  hand, if al_virtualise_fd() fails and returns a negative number, you are
 *  responsible for closing the real file descriptor.
 */
int al_virtualise_fd(int real_fd)
{
   int fd;

   for (fd = 0; fd < n_files; fd++)
      if (!filehooks[fd])
	 goto found_one;

   if (n_files >= MAX_FILES) {
      *allegro_errno = errno = EMFILE;
      return -1;
   }

   n_files++;
   filehooks[fd] = NULL;

   found_one:

   filedupprev[fd] = filedupnext[fd] = fd;

   filedata[fd] = malloc(sizeof(int));

   if (!filedata[fd])
      return -1;

   *(int *)filedata[fd] = real_fd;
   filehooks[fd] = &default_filehooks;

   return fd;
}



/* al_filehook_open:
 *  Replacement for al_open() (which is in turn a replacement for open()).
 *  The file descriptor returned is a virtual descriptor, and the open hooks
 *  will be able to hook the file; otherwise it is functionally equivalent to
 *  al_open(). It is recommended that you use PACKFILEs where possible, but
 *  this function may be useful inside your own file hook functions where you
 *  will need to emulate lseek(). Be careful; this function may call your
 *  hook recursively, and you must detect this.
 */
int al_filehook_open(AL_CONST char *file, int mode)
{
   int i, fd;
   int old_errno;

   for (fd = 0; fd < n_files; fd++)
      if (!filehooks[fd])
	 goto found_one;

   if (n_files >= MAX_FILES) {
      *allegro_errno = errno = EMFILE;
      return -1;
   }

   n_files++;

   found_one:

   /* filehooks[fd] must not be NULL when we call an open hook, in case it
    * calls our function recursively and needs a second fd. For want of
    * anything else to set it to, we set it to point to the default hooks
    * here. Care must be taken in this function to reset it to NULL if
    * failure occurs.
    */
   filehooks[fd] = &default_filehooks;
   filedupprev[fd] = filedupnext[fd] = fd;

   old_errno = *allegro_errno;
   *allegro_errno = errno = 0;

   for (i = 0; i < n_open_hooks; i++) {
      AL_FILEHOOKS *hooks = (*open_hook[i])(file, mode, &filedata[fd]);
      if (hooks) {
	 filehooks[fd] = hooks;
	 return fd;
      }
      if (*allegro_errno) {
	 filehooks[fd] = NULL;
	 return -1;
      }
   }

   *allegro_errno = errno = old_errno;

   filedata[fd] = malloc(sizeof(int));

   if (!filedata[fd])
      return -1;

   i = al_open(file, mode);

   if (i < 0) {
      free(filedata[fd]);
      return -1;
   }

   *(int *)filedata[fd] = i;

   return fd;
}



/* al_filehook_dup:
 *  Replacement for dup(). This will return a second file descriptor you can
 *  use to deal with the same file. The file will not physically be closed
 *  until all existing file descriptors for the file are closed.
 */
int al_filehook_dup(int old_fd)
{
   int new_fd;
   int next_fd;

   for (new_fd = 0; new_fd < n_files; new_fd++)
      if (!filehooks[new_fd])
	 goto found_one;

   if (n_files >= MAX_FILES) {
      *allegro_errno = errno = EMFILE;
      return -1;
   }

   n_files++;

   found_one:

   next_fd = filedupnext[old_fd];

   /* Change this:      old_fd <-> next_fd
    * To this:     old_fd <-> new_fd <-> next_fd
    * The special case with old_fd == next_fd needs no special treatment.
    */
   filedupprev[next_fd] = new_fd;
   filedupnext[old_fd] = new_fd;
   filedupnext[new_fd] = next_fd;
   filedupprev[new_fd] = old_fd;

   filehooks[new_fd] = filehooks[old_fd];
   filedata[new_fd] = filedata[old_fd];

   return new_fd;
}



/* al_filehook_close():
 *  Replacement for close().
 */
int al_filehook_close(int fd)
{
   int rv;

   if ((unsigned)fd >= MAX_FILES || !filehooks[fd]) {
      *allegro_errno = errno = EBADF;
      return -1;
   }

   if (filedupnext[fd] != fd) {
      int next_fd = filedupnext[fd];
      int prev_fd = filedupprev[fd];
      /* Change this: prev_fd <-> fd <-> next_fd
       * To this:        prev_fd <-> next_fd
       * The special case with prev_fd == next_fd needs no special treatment.
       */
      filedupprev[next_fd] = prev_fd;
      filedupnext[prev_fd] = next_fd;

      filehooks[fd] = NULL;
      return 0;
   }

   rv = (*filehooks[fd]->close)(filedata[fd]);

   filehooks[fd] = NULL;

   return rv;
}



/* al_filehook_lseek:
 *  Replacement for lseek().
 */
long al_filehook_lseek(int fd, long offset, int whence)
{
   if ((unsigned)fd >= MAX_FILES || !filehooks[fd]) {
      *allegro_errno = errno = EBADF;
      return -1;
   }

   return (*filehooks[fd]->lseek)(filedata[fd], offset, whence);
}



/* al_filehook_read:
 *  Replacement for read().
 */
long al_filehook_read(int fd, void *buffer, long length)
{
   if ((unsigned)fd >= MAX_FILES || !filehooks[fd]) {
      *allegro_errno = errno = EBADF;
      return -1;
   }

   return (*filehooks[fd]->read)(filedata[fd], buffer, length);
}



/* al_filehook_write:
 *  Replacement for write().
 */
int al_filehook_write(int fd, AL_CONST void *buffer, long count)
{
   if ((unsigned)fd >= MAX_FILES || !filehooks[fd]) {
      *allegro_errno = errno = EBADF;
      return -1;
   }

   return (*filehooks[fd]->write)(filedata[fd], buffer, count);
}

